/* cairot_xlib.c

  X11+xlib version of cairot.

  Copyright (c) 2010 Bill Spitzak (spitzak@gmail.com)

  All code is based on FLTK2.0.

  License is GPL2 or later, plus a linking exception that says you can
  use this code unchanged in any program no matter what the license of
  that program is. However if you modify this code then you must
  include the modified version as source code under this same license
  or (by removing the linking exception) under the GPL or LGPL.

  Please check with www.fltk.org for license details. However the
  intention is to allow this source code to be used for any purpose,
  as long as you don't "close" this source code, without propagating
  the restrictions to other code.
*/

/*
  Current status:

  Only a single window is supported.

  Keystrokes return X11 keysyms rather than a portable value. May not
  be working with XInput and/or returing real UTF-8.
*/

#include "cairot.h"

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <cairo/cairo-xlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Produce less verbose message if connection drops: */
static int io_error_handler(Display* d) {
    fprintf(stderr, "X I/O error\n");
    exit(1);
}

/* Makes it continue on X errors rather than exit, and make readable message: */
static int xerror_handler(Display* d, XErrorEvent* e) {
  char buf1[128], buf2[128];
  sprintf(buf1, "XRequest.%d", e->request_code);
  XGetErrorDatabaseText(d,"",buf1,buf1,buf2,128);
  XGetErrorText(d, e->error_code, buf1, 128);
  fprintf(stderr, "%s: %s 0x%lx\n", buf2, buf1, e->resourceid);
  return 0;
}

static Display* xdisplay = 0;
static int xscreen;
static Atom WM_DELETE_WINDOW;
static Atom WM_PROTOCOLS;
static Atom MOTIF_WM_HINTS;
static Atom XA_TEXT;
static Atom _NET_WM_NAME;
static Atom _NET_WM_ICON_NAME;
static Atom UTF8_STRING;

static Display* open_display() {
  if (xdisplay) return xdisplay;

  XSetIOErrorHandler(io_error_handler);
  XSetErrorHandler(xerror_handler);

  xdisplay = XOpenDisplay(0);
  xscreen = DefaultScreen(xdisplay);

  /* Register all the Atoms we need */
#define MAX_ATOMS 30
  Atom* atom_ptr[MAX_ATOMS];
  const char* names[MAX_ATOMS];
  int i = 0;
#define atom(a,b) atom_ptr[i] = &a; names[i] = b; i++
  atom(	WM_DELETE_WINDOW	, "WM_DELETE_WINDOW");
  atom(	WM_PROTOCOLS		, "WM_PROTOCOLS");
  atom(	MOTIF_WM_HINTS		, "_MOTIF_WM_HINTS");
  atom(	XA_TEXT			, "TEXT");
  atom( _NET_WM_NAME		, "_NET_WM_NAME");
  atom( _NET_WM_ICON_NAME	, "_NET_WM_ICON_NAME");
  atom( UTF8_STRING		, "UTF8_STRING");
#undef atom
  Atom atoms[MAX_ATOMS];
  XInternAtoms(xdisplay, (char**)names, i, 0, atoms);
  for (; i--;) *atom_ptr[i] = atoms[i];

  return xdisplay;
}

/* Local storage for the single window we support */
static int window_w, window_h;
static cairot_window_t redraw_called = 0;

cairot_window_t cairot_create_window(int width, int height)
{

  open_display();

  XSetWindowAttributes attr;
  attr.border_pixel = 0;
  attr.bit_gravity = StaticGravity;
  attr.event_mask =
    ExposureMask | StructureNotifyMask
    | KeyPressMask | KeyReleaseMask | KeymapStateMask | FocusChangeMask
    | ButtonPressMask | ButtonReleaseMask
    | EnterWindowMask | LeaveWindowMask
    | PointerMotionMask;

  window_w = width;
  window_h = height;

  Window window =
    XCreateWindow(xdisplay, RootWindow(xdisplay, xscreen),
                  100, 100, width, height,
                  0, /* borderwidth */
                  DefaultDepth(xdisplay, xscreen),
                  InputOutput,
                  DefaultVisual(xdisplay, xscreen),
                  CWBorderPixel|CWEventMask|CWBitGravity,
                  &attr);

  /* Makes the close button produce an event */
  XChangeProperty(xdisplay, window, WM_PROTOCOLS,
                  XA_ATOM, 32, 0, (unsigned char*)&WM_DELETE_WINDOW, 1);

  /* Make it resizable */
  XSizeHints hints;
  hints.max_width = 2048;
  hints.max_height = 2048;
  hints.win_gravity = StaticGravity;
  hints.flags = PMaxSize|PWinGravity;
  XSetWMNormalHints(xdisplay, window, &hints);

  return window;
}

void cairot_set_window_name(cairot_window_t window, const char* name)
{
  int l = strlen(name);
  XChangeProperty(xdisplay, window, _NET_WM_NAME,
                  UTF8_STRING, 8, 0, (unsigned char*)name, l);
  /* This is only used by old non-UTF-8 window managers: */
  XChangeProperty(xdisplay, window, XA_WM_NAME,
                  XA_STRING, 8, 0, (unsigned char*)name, l);
}

void cairot_set_window_state(cairot_window_t window, unsigned state)
{
  if (state)
    XMapRaised(xdisplay, window);
}

void cairot_get_window_size(cairot_window_t window, int* width, int* height)
{
  // this must be fixed for multiple windows!
  *width = window_w;
  *height = window_h;
}

void cairot_redraw(cairot_window_t window)
{
  // this must be fixed for multiple windows!
  redraw_called = window;
}

static void _draw(cairot_window_t W) {}
static void(*draw)(cairot_window_t) = _draw;
void cairot_set_draw(void(*f)(cairot_window_t)) { draw = f; }

static void _reshape(cairot_window_t W, int x, int y, int w, int h) {}
static void (*reshape)(cairot_window_t, int x, int y, int w, int h) = _reshape;
void cairot_set_reshape(void(*f)(cairot_window_t, int, int, int, int)) { reshape = f; }

static void _close(cairot_window_t W) { exit(0); }
static void(*close)(cairot_window_t) = _close;
void cairot_set_close(void(*f)(cairot_window_t)) { close = f; }

static void _push(cairot_window_t W, int button, int x, int y) {}
static void (*push)(cairot_window_t, int putton, int x, int y) = _push;
void cairot_set_push(void(*f)(cairot_window_t, int, int, int)) { push = f; }

static void _drag(cairot_window_t W, int button, int x, int y) {}
static void (*drag)(cairot_window_t, int putton, int x, int y) = _drag;
void cairot_set_drag(void(*f)(cairot_window_t, int, int, int)) { drag = f; }

static void _release(cairot_window_t W, int button, int x, int y) {}
static void (*release)(cairot_window_t, int putton, int x, int y) = _release;
void cairot_set_release(void(*f)(cairot_window_t, int, int, int)) { release = f; }

static void _motion(cairot_window_t W, int x, int y) {}
static void (*motion)(cairot_window_t, int x, int y) = _motion;
void cairot_set_motion(void(*f)(cairot_window_t, int, int)) { motion = f; }

static void _focus(cairot_window_t W) {}
static void(*focus)(cairot_window_t) = _focus;
void cairot_set_focus(void(*f)(cairot_window_t)) { focus = f; }

static void _keypress(cairot_window_t W, int k, const char* t) {}
static void(*keypress)(cairot_window_t, int, const char*) = _keypress;
void cairot_set_keypress(void(*f)(cairot_window_t, int, const char*)) { keypress = f; }

int cairot_run()
{
  if (!xdisplay) return 0;
  XEvent xevent;
  int sendmotion = 0;
  Window motionwindow;
  int motionX, motionY;
  for (;;) {
    if (sendmotion) {
      sendmotion = 0;
      motion(motionwindow, motionX, motionY);
    }
    if (redraw_called) {
      draw(redraw_called);
      redraw_called = 0;
    }
    XFlush(xdisplay);
    while (XEventsQueued(xdisplay,QueuedAfterReading)) {
      XNextEvent(xdisplay, &xevent);
      Window window = xevent.xany.window;

      switch (xevent.type) {

      case MappingNotify:
	XRefreshKeyboardMapping((XMappingEvent*)&xevent.xmapping);
	break;

      case ClientMessage:
	if (xevent.xclient.data.l[0] == WM_DELETE_WINDOW)
	  close(window);
	break;

      case ConfigureNotify:
      case MapNotify: {
	/* event data is useless due to WM bugs, do a query instead */
	XWindowAttributes actual;
	XGetWindowAttributes(xdisplay, window, &actual);
	int W = actual.width; int H = actual.height;
	Window junk; int X, Y;
	XTranslateCoordinates(xdisplay, window, actual.root,
			      0, 0, &X, &Y, &junk);
	// Needs to do redraw if it gets smaller as X sends no expose event:
	if ((W < window_w && H <= window_h) || (W <= window_w && H < window_h))
	  cairot_redraw(window);
	window_w = W; window_h = H;
	reshape(window, X, Y, W, H);
	break;}

      case Expose:
      case GraphicsExpose:
	cairot_redraw(window);
	break;

      case ButtonPress:
	push(window, xevent.xbutton.button, xevent.xbutton.x, xevent.xbutton.y);
	break;

      case MotionNotify:
	if (xevent.xbutton.state & 0x7F00) {
	  sendmotion = 0;
	  int button = 1;
	  while (!(xevent.xbutton.state & (0x80<<button))) ++button;
	  drag(window, button, xevent.xbutton.x, xevent.xbutton.y);
	} else {
	  sendmotion = 1;
	  motionwindow = window;
	  motionX = xevent.xbutton.x;
	  motionY = xevent.xbutton.y;
	}
	break;

      case EnterNotify:
	if (!(xevent.xcrossing.state & 0x7F00)) {
	  sendmotion = 1;
	  motionwindow = window;
	  motionX = xevent.xbutton.x;
	  motionY = xevent.xbutton.y;
	}
	break;

      case LeaveNotify:
	if (!(xevent.xcrossing.state & 0x7F00)) {
	  sendmotion = 1;
	  motionwindow = 0;
	  motionX = xevent.xbutton.x_root;
	  motionY = xevent.xbutton.y_root;
	}
	break;

      case ButtonRelease:
	release(window,
		xevent.xbutton.button, xevent.xbutton.x, xevent.xbutton.y);
	break;

      case FocusIn:
	focus(window);
	break;

      case FocusOut:
	focus(0);
	break;

      case KeyPress: {
	KeySym keysym;
	int length;
	char buffer[20];
	length = XLookupString(&(xevent.xkey), buffer, sizeof(buffer)-1,
			       &keysym, 0);
	buffer[length] = 0;
	keypress(window, keysym, buffer);
	break;}
      }
    }
  }
}

/* Returns a cairo_t for drawing into window, reusing as much as possible. */
cairo_t* cairo_for(cairot_window_t window)
{
  static Window pwindow = 0;
  static cairo_surface_t* surface = 0;
  static cairo_t* cr = 0;
  static int pw, ph;
  int W, H; cairot_get_window_size(window, &W, &H);

  /* Spew and reset any previous error */
  if (cr) {
    cairo_status_t cstatus = cairo_status(cr);
    if (cstatus) {
      printf("Cairo: %s", cairo_status_to_string(cstatus));
      pwindow = 0;
    }
  }
  if (pwindow != window) {
    if (cr) {
      cairo_destroy(cr);
      cairo_surface_destroy(surface);
    }
    surface =
      cairo_xlib_surface_create(xdisplay, window,
				DefaultVisual(xdisplay, xscreen),
				W, H);
    cr = cairo_create(surface);
    pwindow = window;
    pw = W;
    ph = H;
  } else if (pw != W || ph != H) {
    cairo_xlib_surface_set_size(surface, W, H);
    pw = W;
    ph = H;
  }
  return cr;
}

/* end of cairot_xlib.c */
